# fptokens py.test configuration
import pytest
import fptokens as fpt


@pytest.fixture(scope='session')
def tokens():
    t1 = fpt.Token('$token$')
    t2 = fpt.Token('$token$')
    t3 = fpt.Token('>token>', escape='>')
    t4 = fpt.Token('$alt_token$')
    return (t1, t2, t3, t4)


@pytest.fixture(scope='session')
def filename():
    fp = fpt.Filename(root='',
                      folders=['$token$', 'subfolder'],
                      base=['this', 'file', '$token2$'])
    return fp



import os
import re
import attr

from path import Path

__version__ = '0.1.0'
__all__ = []


@attr.s(cmp=False, hash=False, repr=False)
class Token(object):
    """
    Token object for use in :class:`~fptokens.Filename`.

    :param name: Token name
    :type name: str
    :param escape: Escape character, default: ``$``
    :type escape: str
    """
    _name = attr.ib()
    escape = attr.ib(default='$')

    def __attrs_post_init__(self):
        self._name = self.__convert_token(self._name)

    @property
    def name(self):
        """
        Return the token name.

        :return: Token name
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, value):
        """
        Given a ``value``, strip the value of the escape character and set
        the token name.

        :param value: New token name
        :type value: str
        """
        self._name = self.__convert_token(value)

    @property
    def token(self):
        """
        Return the full token including the escape characters either side.

        :return: Token
        :rtype: str
        """
        return '{0}{1}{0}'.format(self.escape, self.name)

    def __convert_token(self, token):
        pattern = r'\{esc}(\w+)\{esc}'.format(esc=self.escape)
        match = re.match(pattern, token)
        if not match:
            raise ValueError('Invalid token')
        return match.group(1)

    def __key(self):
        return (self.name, self.escape)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.__key() == other.__key()
        return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(self.__key())

    def __repr__(self):
        return '<Token: {0}>'.format(self.token)


@attr.s(cmp=False, hash=False, repr=False)
class Filename(object):
    """
    Filenames with support for tokens.

    :param root: Root location
    :type root: str
    :param folders: Folder names, attribute supports tokens
    :type folders: list of str
    :param base: Basename of the file, attribute supports tokens
    :type base: list of str
    :param separator: Separator for basename elements
    :type separator: str
    :param extension: Filename extension
    :type extension: str
    :param escape: Escape character for tokens, default: ``$``
    :type escape: str
    """
    _root = attr.ib()
    folders = attr.ib(default=[],
                      validator=attr.validators.instance_of(list))
    base = attr.ib(default=[],
                   validator=attr.validators.instance_of(list))
    separator = attr.ib(default='_',
                        validator=attr.validators.instance_of(str))
    extension = attr.ib(default='jpg',
                        validator=attr.validators.instance_of(str))
    escape = attr.ib(default='$')

    def __attrs_post_init__(self):
        self._root = Path(self._root)

    @property
    def root(self):
        """
        Return the filename's root location.

        :return: Root location
        :rtype: :class:`~path.Path`
        """
        return self._root

    @root.setter
    def root(self, value):
        """
        Given a ``value``, convert the value to a :class:`~path.Path` and set
        the filename's root location.

        :param value: New root location
        :type value: str
        """
        self._root = Path(value)

    @property
    def dirname(self):
        """
        Return the filename's location.

        :return: Dirname
        :rtype: str
        """
        return self.root / os.path.sep.join(
            [str(folder) for folder in self.folders])

    @property
    def basename(self):
        """
        Return the filename's basename.

        :return: Basename
        :rtype: str
        """
        return '{0}.{1}'.format(self.separator.join(
            [str(base) for base in self.base]),
            self.extension)

    @property
    def tokens(self):
        """
        Return a list of all tokens in ``folders`` and ``base``.

        :return: List of tokens
        :rtype: list
        """
        folders = set([token for token in self.folders
                       if isinstance(token, Token)])
        basename = set([token for token in self.base
                        if isinstance(token, Token)])
        return [token for token in folders.union(basename)]

    @property
    def abspath(self):
        """
        Return the filename's full absolute path.

        :return: Absolute path
        :rtype: :class:`~path.Path`
        """
        return (self.dirname / self.basename).abspath()

    def make(self):
        """Create the filename's location if it does not exist."""
        if any([token for token in self.folders if isinstance(token, Token)]):
            raise ValueError('Replace tokens with string values '
                             'to create folders')
        self.dirname.makedirs_p()

    def parse(self):
        """
        Parse the filename's ``folders`` and ``base`` attributes, detect
        components that match the token pattern and replace these with
        :class:`fptokens.Token` objects.
        """
        for base_components in (self.folders, self.base):
            for component_idx, component in enumerate(base_components):
                if isinstance(component, Token):
                    continue
                pattern = r'(\{esc}\w+\{esc})'.format(esc=self.escape)
                tokens = re.findall(pattern, component)
                if tokens:
                    if len(tokens) > 1:
                        raise ValueError('Limit: one token per component.')
                    token = tokens[0]
                    base_components[component_idx] = Token(name=token,
                                                           escape=self.escape)

    def __key(self):
        return (self.root, self.folders, self.basename,
                self.separator, self.extension)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.__key() == other.__key()
        return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(self.__key())

    def __repr__(self):
        return '<Filename: {0}>'.format(self.abspath)



# fptokens tests
import pytest
from path import Path
import fptokens as fpt


def test_token_init():
    token = fpt.Token(name='$token$')
    assert token
    assert token.name == 'token'
    assert token.token == '$token$'


def test_token_change():
    token = fpt.Token(name='$token$')
    token.name = '$replace_token$'
    assert token.name == 'replace_token'
    assert token.token == '$replace_token$'


def test_token_invalid_init():
    with pytest.raises(ValueError):
        fpt.Token(name='>token>')
    with pytest.raises(ValueError):
        fpt.Token(name='$token$', escape='>')


def test_tokens(tokens):
    assert tokens[0] == tokens[1]
    assert tokens[0] != tokens[2]
    assert tokens[0] != tokens[3]
    assert tokens[0].name == 'token'
    assert tokens[0] != True


def test_filename_init(tmpdir):
    fp = fpt.Filename(root=tmpdir,
                      folders=['this', 'folder'],
                      base=['this', 'file'],
                      separator='_',
                      extension='png')
    assert fp

def test_filename_equality(tmpdir):
    fp1 = fpt.Filename(root=tmpdir,
                       folders=['$token$', 'subfolder'],
                       base=['this', 'file', '$token2$'])
    fp2 = fpt.Filename(root=tmpdir,
                       folders=['$token$', 'subfolder'],
                       base=['this', 'file', '$token2$'])
    assert fp1 == fp2
    fp2.folders.append('another_subfolder')
    assert fp1 != fp2
    assert fp1 != True


def test_root(filename, tmpdir):
    filename.root = tmpdir
    assert isinstance(filename.root, Path)
    assert filename.root.exists()


def test_dirname(filename):
    assert isinstance(filename.dirname, Path)
    assert not filename.dirname.exists()
    filename.make()
    assert filename.dirname.exists()


def test_basename(filename):
    assert filename.basename == 'this_file_$token2$.jpg'


def test_abspath(filename):
    assert isinstance(filename.dirname, Path)
    assert filename.abspath == (filename.dirname / filename.basename).abspath()
    assert not filename.abspath.exists()


def test_parsing(filename):
    filename.parse()
    assert len(filename.tokens) == 2
    assert all([token for token in filename.tokens if isinstance(token,
        fpt.Token)])


def test_second_parsing(filename):
    filename.parse()
    assert len(filename.tokens) == 2
    assert all([token for token in filename.tokens if isinstance(token,
        fpt.Token)])


def test_make_invalid(filename):
    with pytest.raises(ValueError):
        filename.make()


def test_parsing_invalid(tmpdir):
    fp = fpt.Filename(root=tmpdir,
                      folders=['$token$$token$', 'subfolder'],
                      base=['this', 'file', '$token2$'])
    with pytest.raises(ValueError):
        fp.parse()
